(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Comp IMPORT TValue, TOperation, Val, Opr;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TObjDesc = RECORD
    i: CARDINAL;
    v: TValue;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Add();
VAR i: CARDINAL;
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].i, id ) = 0 THEN Err( 'Constant already declared' ) END;
    INC( i );
  END;
  IF cnt >= cntMax THEN Err( 'Too many declarations' ) END;
  IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
  objs[ cnt ].i := id;
  objs[ cnt ].v := num;
  INC( cnt );
END Add;

PROCEDURE Get( VAR x: TValue );
VAR i: CARDINAL;
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].i, id ) = 0 THEN x := objs[ i ].v; RETURN END;
    INC( i );
  END;
  Err( 'Identifier undeclared' );
END Get;

PROCEDURE Expression();

  PROCEDURE Factor();
  VAR x: TValue;
  BEGIN
    IF sym = sNum THEN
      IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
      Val( TValue(num) ); Read();
    ELSIF sym = sId THEN
      Get( x ); Val( x ); Read();
    ELSIF sym = sLPar THEN
      Read(); Expression();
      IF sym = sRPar THEN Read() ELSE Err( 'Expected operator or ")"' ) END;
    ELSE
      Err( 'Expected number, identifier or "("' );
    END;
  END Factor;

  PROCEDURE Term();
  VAR op: TSymbol;
  BEGIN
    Factor();
    WHILE ( sym = sStar ) OR ( sym = sSlash ) DO
      op := sym; Read(); Factor();
      IF op = sStar THEN Opr( oMul ) ELSE Opr( oDiv ) END;
    END;
  END Term;

VAR op: TSymbol;
BEGIN
  IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
    op := sym; Read(); Term();
    IF op = sMinus THEN Opr( oNeg ) END;
  ELSE
    Term();
  END;
  WHILE ( sym = sPlus ) OR ( sym = sMinus ) DO
    op := sym; Read(); Term();
    IF op = sPlus THEN Opr( oAdd ) ELSE Opr( oSub ) END;
  END;
END Expression;

PROCEDURE Statement();
BEGIN
  IF sym = sId THEN       (* constant declaration *)
    Read(); IF sym # sEql THEN Err( 'Expected "="' ) END;
    Read(); IF sym # sNum THEN Err( 'Expected number' ) END;
    Add(); KeepId(); Read();
  ELSIF sym = sExcl THEN  (* expression output *)
    Read(); Expression(); Opr( oWrt );
  END;
END Statement;

PROCEDURE Run();
BEGIN
  Read(); Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
  IF sym # sDot THEN Err( 'Expected "."' ) END;
END Run;

BEGIN
  cnt := 0;
END Prsr.
